""" tools for generating launch files for ROS """
import yaml

from lxml import etree
from util import convert_ros


def generate_leg_graph_launch(key, leg):

    launch = etree.Element('launch')

    etree.SubElement(
        launch,
        'node',
        pkg='schpin_stg',
        type='leg_graph_generator',
        name='{}_leg_graph_generator'.format(key)
    )
    rosparam = etree.SubElement(launch, 'rosparam')
    rosparam.text = yaml.dump(leg)

    return launch


def generate_robot_description():

    launch = etree.Element("launch")

    etree.SubElement(
        launch,
        "param",
        name="robot_description",
        textfile="$(find schpin_tote)/src/tote.urdf"
    )

    return launch


# TODO: lxml should be dependency, so every other module, maybe grep-check this?
def generate_servo_pwm_node(xml_root, tf_prefix, name, segment):
    if 'joint' not in segment:
        return
    joint = etree.SubElement(
        xml_root,
        "node",
        pkg="schpin_dev",
        type="pwm_servo_joint.py",
        name=name,
        respawn="true"
    )

    params = {
        'angle_min': segment['joint']['angle']['min'],
        'angle_max': segment['joint']['angle']['max'],
        'pulse_min': segment['joint']['pulse']['min'],
        'pulse_max': segment['joint']['pulse']['max'],
        "velocity": segment['joint']['max_velocity'],
        "update_freq": 50,  # TODO: fix this
        "reverse": segment['joint']['reverse'],
        "controller_topic": "/hw/controller_topic",
        "joint_state_topic": "joint_states",
        'channel': 0,
        'default_post': 0,
        'tf_frame': tf_prefix + name
    }

    for key, val in params.items():
        etree.SubElement(joint, "param", name=key, value=str(convert_ros(val)))


def generate_dev_layer(cfg):

    launch = etree.Element('launch')

    etree.SubElement(
        launch,
        "include",
        file="$(find schpin_tote)/launch/robot_description.launch"
    )

    etree.SubElement(
        launch,
        "node",
        name="robot_state_publisher",
        pkg="robot_state_publisher",
        type="robot_state_publisher"
    )

    js_pub = etree.SubElement(
        launch,
        "node",
        name="joint_state_publisher",
        pkg="joint_state_publisher",
        type="joint_state_publisher",
    )

    js_params = etree.SubElement(js_pub, "rosparam")
    js_params.text = yaml.dump(
        {
            "source_list":
            ["/dev/leg/" + x + "/joint_states" for x in cfg['legs'].keys()] +
            ["/dev/gimbal/joint_states"]
        }
    )

    dev_group = etree.SubElement(launch, "group", ns="dev")

    leg_group = etree.SubElement(dev_group, "group", ns="leg")

    for key, l_cfg in cfg['legs'].items():

        lgroup = etree.SubElement(leg_group, "group", ns=key)
        iterate_chain( l_cfg['chain_root'], l_cfg['segments'], lambda name, segment: generate_servo_pwm_node(lgroup, "leg/" + key + "/", name , segment))

    ggroup = etree.SubElement(dev_group, "group", ns="gimbal")

    iterate_chain(
        cfg['gimbal']['chain_root'], cfg['gimbal']['segments'],
        lambda name, segment: generate_servo_pwm_node(ggroup, "cam/", name, segment)
    )

    return launch


def iterate_chain(name, segments, func):
    #TODO: use this elsewhere
    func(name, segments[name])
    if 'next_segment' in segments[name]:
        iterate_chain(segments[name]['next_segment'], segments, func)


def generate_launch_files(cfg):

    files = {
        "robot_description.launch": generate_robot_description(),
        "dev_layer.launch": generate_dev_layer(cfg)
    }

    for key in cfg['legs'].keys():
        files["{}_leg_grpah_generator.launch".format(key)
              ] = generate_leg_graph_launch(key, cfg['legs'][key])

    for filename, element in files.items():
        tree = etree.ElementTree(element)

        tree.write("../launch/" + filename, pretty_print=True)
